/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TabLang extension for Mozilla Firefox.
 *
 * The Initial Developer of the Original Code is Andrey Lokot.
 * Portions created by the Initial Developer are Copyright (C) 2010-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrey Lokot <andreylokot@yandex.ru>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var EXPORTED_SYMBOLS = ["TablangKbSwitcherLinux"];

Components.utils.import("resource://gre/modules/ctypes.jsm");

const Bool = ctypes.int;
const False = 0;
const True = 1;

const Status = ctypes.int;
const Success = 0;

const DisplayPtr = ctypes.voidptr_t;

const KeyCode = ctypes.uint8_t;
const Atom = ctypes.unsigned_long;

const XkbStateRec = ctypes.StructType("XkbStateRec", [
	{"group": ctypes.unsigned_char},
	{"locked_group": ctypes.unsigned_char},
	{"base_group": ctypes.unsigned_short},
	{"latched_group": ctypes.unsigned_short},
	{"mods": ctypes.unsigned_char},
	{"base_mods": ctypes.unsigned_char},
	{"latched_mods": ctypes.unsigned_char},
	{"locked_mods": ctypes.unsigned_char},
	{"compat_state": ctypes.unsigned_char},
	{"grab_mods": ctypes.unsigned_char},
	{"compat_grab_mods": ctypes.unsigned_char},
	{"lookup_mods": ctypes.unsigned_char},
	{"compat_lookup_mods": ctypes.unsigned_char},
	{"ptr_buttons": ctypes.unsigned_short}
]);

const XkbNumVirtualMods = 16;
const XkbNumIndicators = 32;
const XkbNumKbdGroups = 4;

const XkbNamesRec = ctypes.StructType("XkbNamesRec", [
	{"keycodes": Atom},
	{"geometry": Atom},
	{"symbols": Atom},
	{"types": Atom},
	{"compat": Atom},
	{"vmods": Atom.array(XkbNumVirtualMods)},
	{"indicators": Atom.array(XkbNumIndicators)},
	{"groups": Atom.array(XkbNumKbdGroups)},
	{"keys": ctypes.voidptr_t},
	{"key_aliases": ctypes.voidptr_t},
	{"radio_groups": Atom.ptr},
	{"phys_symbols": Atom},
	{"num_keys": ctypes.unsigned_char},
	{"num_key_aliases": ctypes.unsigned_char},
	{"num_rg": ctypes.unsigned_short},
]);

const XkbDescRec = ctypes.StructType("XkbDescRec", [
	{"dpy": DisplayPtr},
	{"flags": ctypes.unsigned_short},
	{"device_spec": ctypes.unsigned_short},
	{"min_key_code": KeyCode},
	{"max_key_code": KeyCode},
	{"ctrls": ctypes.voidptr_t},
	{"server": ctypes.voidptr_t},
	{"map": ctypes.voidptr_t},
	{"indicators": ctypes.voidptr_t},
	{"names": XkbNamesRec.ptr},
	{"compat": ctypes.voidptr_t},
	{"geom": ctypes.voidptr_t}
]);

const XkbMajorVersion = 1;
const XkbMinorVersion = 0;

const XkbUseCoreKbd = ctypes.unsigned_int("0x0100");
const XkbGroupNamesMask = ctypes.unsigned_int("0x1000");

function TablangKbSwitcherLinux() {
	let libgdk = ctypes.open("libgdk-x11-2.0.so.0");
	this.display = DisplayPtr(libgdk.declare("gdk_display", DisplayPtr));
	libgdk.close()
	this.libX11 = ctypes.open("libX11.so.6");
	this.xkb = {
		XkbLibraryVersion: this.libX11.declare("XkbLibraryVersion", ctypes.default_abi,
			Bool, ctypes.int.ptr, ctypes.int.ptr),
		XkbQueryExtension: this.libX11.declare("XkbQueryExtension", ctypes.default_abi,
			Bool, DisplayPtr, ctypes.int.ptr, ctypes.int.ptr, ctypes.int.ptr, ctypes.int.ptr, ctypes.int.ptr),
		XkbAllocKeyboard: this.libX11.declare("XkbAllocKeyboard", ctypes.default_abi,
			XkbDescRec.ptr),
		XkbFreeKeyboard: this.libX11.declare("XkbFreeKeyboard", ctypes.default_abi,
			ctypes.void_t, XkbDescRec.ptr, ctypes.unsigned_int, Bool),
		XkbGetNames: this.libX11.declare("XkbGetNames", ctypes.default_abi,
			Status, DisplayPtr, ctypes.unsigned_int, XkbDescRec.ptr),
		XGetAtomName: this.libX11.declare("XGetAtomName", ctypes.default_abi,
			ctypes.char.ptr, DisplayPtr, Atom),
		XFree: this.libX11.declare("XFree", ctypes.default_abi,
			ctypes.void_t, ctypes.voidptr_t),
		XkbGetState: this.libX11.declare("XkbGetState", ctypes.default_abi,
			Status, DisplayPtr, ctypes.unsigned_int, XkbStateRec.ptr),
		XkbLockGroup: this.libX11.declare("XkbLockGroup", ctypes.default_abi,
			Bool, DisplayPtr, ctypes.unsigned_int, ctypes.unsigned_int)
	};
	let majorVersion = ctypes.int(XkbMajorVersion);
	let minorVersion = ctypes.int(XkbMinorVersion);
	let opcode = ctypes.int();
	let event = ctypes.int();
	let error = ctypes.int();
	if (this.xkb.XkbLibraryVersion(majorVersion.address(), minorVersion.address()) != True) {
		this.libX11.close();
		throw new Error("Xkb version check (XkbLibraryVersion) failed");
	}
	if (this.xkb.XkbQueryExtension(this.display, opcode.address(), event.address(), error.address(),
			majorVersion.address(), minorVersion.address()) != True) {
		this.libX11.close();
		throw new Error("XkbQueryExtension failed");
	}
	this.updateNamesCache();
}

TablangKbSwitcherLinux.prototype = {
	getCurrentLang: function TablangKbSwitcherLinux_getCurrentLang() {
		let state = XkbStateRec();
		if (this.xkb.XkbGetState(this.display, XkbUseCoreKbd, state.address()) != Success) {
			throw new Error("XkbGetState failed");
		}
		return (new Number(state.group)).toString();
	},
	setCurrentLang: function TablangKbSwitcherLinux_setCurrentLang(newLang) {
		if (this.xkb.XkbLockGroup(this.display, XkbUseCoreKbd, parseInt(newLang)) != True) {
			throw new Error("XkbLockGroup failed with Xkb group = " + group);
		}
	},
	getLangList: function TablangKbSwitcherLinux_getLangList() {
		this.updateNamesCache();
		let langList = [];
		for (let group = 0; group < this.groupNames.length; group++) {
			if (this.groupNames[group] !== null) {
				langList.push(group.toString());
			}
		}
		return langList;
	},
	getLangName: function TablangKbSwitcherLinux_getLangName(lang) {
		this.updateNamesCache();
		let group = parseInt(lang);
		return this.groupNames[group];
	},
	updateNamesCache: function TablangKbSwitcherLinux_updateNamesCache() {
		let keyboard = XkbDescRec.ptr(this.xkb.XkbAllocKeyboard());
		if (keyboard.isNull()) {
			throw new Error("XkbAllocKeyboard failed");
		}
		let groupNames = new Array(XkbNumKbdGroups);
		try {
			if (this.xkb.XkbGetNames(this.display, XkbGroupNamesMask, keyboard) != Success) {
				throw new Error("XkbGetNames failed");
			}
			for (let group = 0; group < XkbNumKbdGroups; group++) {
				let groupNameAtom = Atom(keyboard.contents.names.contents.groups[group])
				if (groupNameAtom.value != 0) {
					let groupName = ctypes.char.ptr(this.xkb.XGetAtomName(this.display, groupNameAtom));
					if (groupName.isNull()) {
						throw new Error("XGetAtomName failed for group = " + group);
					}
					groupNames[group] = groupName.readString() + " [group " + group + "]";
					this.xkb.XFree(groupName);
				}
				else {
					groupNames[group] = "[group " + group + "]";
				}
			}
			this.groupNames = groupNames;
		}
		finally {
			this.xkb.XkbFreeKeyboard(keyboard, 0, True);
		}
	}
};

